﻿using System;
using System.Threading;

namespace CornerCases
{
    class RaisingEvents
    {
        public event EventHandler BadEvent1;
        public event EventHandler BadEvent2;
        public event EventHandler StaleEvent;
        public event EventHandler LockedEvent;
        public event EventHandler OldCarefulEvent;
        public event EventHandler NewCarefulEvent;
        public event EventHandler NewHelpedEvent;
        public event EventHandler SafeEvent1 = delegate {};
        public event EventHandler SafeEvent2 = (sender, args) => { };

        protected void OnBadEvent1()
        {
            BadEvent1(this, EventArgs.Empty);
        }

        protected void OnBadEvent2()
        {
            if (BadEvent2 != null)
            {
                BadEvent2(this, EventArgs.Empty);
            }
        }

        protected void OnStaleEvent()
        {
            EventHandler handler = StaleEvent;
            if (handler != null)
            {
                handler(this, EventArgs.Empty);
            }
        }

        protected void OnLockedEvent()
        {
            lock (this)
            {
                EventHandler handler = LockedEvent;
                if (handler != null)
                {
                    handler(this, EventArgs.Empty);
                }
            }
        }

        protected void OnOldCarefulEvent()
        {
            EventHandler handler;
            lock (this)
            {
                handler = OldCarefulEvent;
            }
            if (handler != null)
            {
                handler(this, EventArgs.Empty);
            }
        }

        protected void OnNewCarefulEvent()
        {
            EventHandler handler = Interlocked.CompareExchange(ref NewCarefulEvent, null, null);
            if (handler != null)
            {
                handler(this, EventArgs.Empty);
            }
        }

        protected void OnNewHelpedEvent()
        {
            EventHelpers.RaiseEvent(ref NewHelpedEvent, this, EventArgs.Empty);
        }

        protected void OnSafeEvent1()
        {
            SafeEvent1(this, EventArgs.Empty);
        }

        protected void OnSafeEvent2()
        {
            SafeEvent2(this, EventArgs.Empty);
        }

        static class EventHelpers
        {
            internal static void RaiseEvent(ref EventHandler handler, object sender, EventArgs args)
            {
                EventHandler realHandler = Interlocked.CompareExchange(ref handler, null, null);
                if (realHandler != null)
                {
                    realHandler(sender, args);
                }
            }
        }
    }
}
